from pyvgmdb.utils import SafeEtree
from pyvgmdb.notes import Notes
import requests
from lxml import etree
import warnings
import os
import re


class Album:
    def __init__(self, album_id, proxies=None, headers=None):
        """
        :param album_id: e.g. 25781, then we will parse: https://vgmdb.net/album/25781
        :param proxies: if you need proxies to visit https://vgmdb.net, please set this param
        :param headers: not necessary
        """
        self.album_id = str(album_id)
        if int(album_id) > 0:
            self.album_url = "https://vgmdb.net/album/" + self.album_id
            self.proxies = {'http': f'http://127.0.0.1:{proxies}', 'https': f'http://127.0.0.1:{proxies}'}
            self.headers = headers
            self.tree, self.page_text = self.getPageEtree()
            if self.tree is not None:
                self.album_info = {
                    'album_id': self.album_id,
                    'album_basic_info': self.getAlbumBasicInfo(),
                    'album_credits': self.getCredits(),
                    'track_list_info': self.getTrackList(),
                    'album_notes': self.getNotes(),
                }
                if self.album_info['track_list_info']:
                    self.parseNotes()

    def getPageEtree(self):
        args = {'url': self.album_url}
        if self.proxies:
            args['proxies'] = self.proxies
        if self.headers:
            args['headers'] = self.headers
        response = requests.get(**args)
        page_text = response.text
        if response.status_code == 404 or "This album could not be displayed" in page_text:
            warnings.warn(f"can not find album with id {self.album_id}")
            return None, None
        tree = SafeEtree(etree.HTML(page_text))
        return tree, page_text

    def getAlbumBasicInfo(self):
        album_basic_info = {
            'album_name_en': self.tree.xpath('album_name_en')[0],
            'album_name_ja': self.tree.xpath('album_name_ja')[0],
            'album_name_ja_latn': self.tree.xpath('album_name_ja_latn')[0],
        }
        album_info_tr = self.tree.xpath('album_info_tr')
        for tr in album_info_tr:
            key_name_td_list = tr.xpath('./td')
            if key_name_td_list:
                key_name = key_name_td_list[0].xpath('.//text()')[0]
                if len(tr.xpath('./td')) >= 2:
                    album_basic_info[key_name] = tr.xpath('./td')[1].xpath('.//text()')
        if len(album_basic_info) == 0:
            warnings.warn(f"can not find any basic info in album with id {self.album_id}")
        return album_basic_info

    def getCredits(self):
        album_credits = {}
        if "collapse_credits" not in self.page_text:
            warnings.warn(f"can not find Credits info in album with id {self.album_id}")
            return album_credits
        for tr in self.tree.xpath('album_credits'):
            key_name = tr.xpath('./td')[0].xpath('.//span[@lang="en"]/text()')[0]
            album_credits[key_name] = []
            # person with no href
            if tr.xpath('./td')[1].xpath('./text()'):
                album_credits[key_name].append(tr.xpath('./td')[1].xpath('./text()')[0])
            # person with href
            for a in tr.xpath('./td')[1].xpath('./a'):
                href = a.xpath('./@href')[0]
                artist_id = str(href).split('/')[-1]
                # person with only en name
                if a.xpath('./text()'):
                    name = a.xpath('./text()')[0]
                    album_credits[key_name].append({'name': name, 'artist_id': artist_id})
                # person with many types of name
                elif a.xpath('./span[@lang="en"]/text()'):
                    name = a.xpath('./span[@lang="en"]/text()')[0]
                    album_credits[key_name].append({'name': name, 'artist_id': artist_id})
                else:
                    raise Exception(f"can not find name of person who is {key_name} "
                                    f"in album with id {self.album_id}")
        return album_credits

    def getTrackList(self):
        album_track_list_info = []
        if "No tracklist found" in "".join(self.tree.etree.xpath('//*[@id="tracklist"]//text()')):
            warnings.warn(f"No tracklist found in album with id {self.album_id}")
            return album_track_list_info
        for span in self.tree.xpath('track_list_span'):
            tl_id = span.xpath('./@id')[0]
            try:
                lang = self.tree.etree.xpath(f'//*[@id="tlnav"]/li/a[@rel="{tl_id}"]/text()')[0]
            except IndexError:
                raise Exception(f"can not find language type of track_list in album with id {self.album_id}")
            disc_name_list = span.xpath('./span/b/text()')
            disc_info_list = span.xpath('./table')
            assert len(disc_name_list) == len(disc_info_list), f"num of disc is not sure, " \
                                                               f"we find following disc names: {disc_name_list}" \
                                                               f"but there are {len(disc_info_list)} disc info tables" \
                                                               f"please check album with id {self.album_id}"
            # for num of disc
            album_tl_info = {
                'info_language': lang,
                'track_list': [],
            }
            for disc_name, tb in zip(disc_name_list, disc_info_list):
                disc_track_list = {
                    'disc_name': disc_name,
                    'disc_track_list': [],
                }
                for tr in tb.xpath('./tr[@class="rolebit"]'):
                    name = tr.xpath('./td[2]/text()')[0]
                    name = name.replace('\n', '').replace('\r', '').replace('\t', '')
                    song_length_text = tr.xpath('./td[3]/span[@class="time"]/text()')
                    if song_length_text:
                        song_length = song_length_text[0]
                    else:
                        song_length = None
                    disc_track_list['disc_track_list'].append({'name': name, 'time': song_length})
                album_tl_info['track_list'].append(disc_track_list)
            album_track_list_info.append(album_tl_info)
        return album_track_list_info

    def getNotes(self):
        match = re.search(r'(?<=\sid="notes">).*?(?=</div>)', self.page_text)
        if match is None:
            warnings.warn(f"no note found in album with id {self.album_id}")
            return
        else:
            return match.group()

    def parseNotes(self):
        note = Notes(self.album_info)
        self.album_info['album_parsed_notes'] = note.album_note_parse

    def save(self, fp):
        with open(os.path.join(fp + str(self.album_id) + ".json"), 'w', encoding="utf-8") as fs:
            fs.write(json.dumps(self.album_info))


if __name__ == '__main__':

    import numpy as np
    import time
    import json

    test_id = list(range(40000, 100000))
    np.random.shuffle(test_id)
    bad_example = []
    for k in bad_example:
        print("current album id：", k)
        Album(k, proxies=7890).save("./test_note/")
        time.sleep(12)

    for k in test_id[:30]:
        if k in bad_example:
            continue
        print("current album id：", k)
        try:
            Album(k, proxies=7890).save("./test_note/")
            print("album id", k, "success")
        except Exception:
            bad_example.append(k)
            print(":( "*20)
            print("album id", k, "failure!")
        time.sleep(8 + np.random.randint(10))
    print(bad_example)



from selenium.webdriver import Chrome
from thread_tools import res_pool_parallel

driver = Chrome()


def open_urls_in_Chrome(url_list):
    global driver
    res_pool_parallel(
        driver.execute_script,
        [(f"window.open('{url}')",) for url in url_list]
    )


if __name__ == '__main__':
    """ test """
    # open_urls_in_Chrome(["https://vgmdb.net/artist/"+str(num) for num in range(100, 130)])



import re

pattern_track = r"(\(|\[)?(M|([aA]ll\s)?[tT][rR][aA][cC][kK][sS]?|[dD]isc|DISC|#|\s|-|\.|,|#)*\d+" + \
                r"(\d|\sto\s|\s|\.|~|\(|\)|\[|\]|#|-|,)+"
pattern_disc = r"disc(\s|\.|-|\)|\]|#)*\d+"
pattern_semicolon = r":|："

patterns_credit = {
    'Composer': ['Composers', 'Composer', 'Composition', 'Composed by', 'Composed', 'Written by',
                 'Music by', r'^\s*Music\s*:'],
    'Arranger': ['Arrangers', 'Arranger', 'Arrangements', 'Arrangement', 'Arranged by', 'Arranged'],
    'Vocal': ['Vocal', 'Vocals', 'Singer', 'Singers'],
    'Lyrics': ['Lyricist', 'Lyricists', 'Lyrics'],
}


# Remove prefix and suffix matching chars
def rm_presuffix(pattern, text):
    ptn = f"[^('+({pattern})+')].*[^('+({pattern})+')]"
    match = re.search(ptn, text)
    if match is not None:
        return re.search(ptn, text).group()
    else:
        return ''


def split_by_re(pattern, text, mod="full", absolut_ptn_len=None, flags=None):
    """e.g.
    1.
    text = "x1 x2 x3",  pattern = r"x" , mod="prefix" , absolut_ptn_len=1  -> return:
    ["x1", "x2", "x3"]

    2.
    text = "x1 x2 x3",  pattern = r"x" , mod="full"  -> return:
    ["x", "1 ", "x", "2 ", "x", "3"]

    3.
    text = "x1 x2 x3",  pattern = r"x" , mod="suffix"  -> return:
    ["x", "1 x", "2 x", 3"]

    absolut_ptn_len: used only when mod="prefix". This param is your pattern's certain length, if you are not sure,
    you can set a rough number, but may cause come mistake.
    what's more, you can set: `flags=re.I` to match both uppercase and lowercase
    """
    assert mod in ["full", "prefix", "suffix"]
    if mod == "prefix":
        assert absolut_ptn_len is not None and 1 <= absolut_ptn_len <= len(
            text), "if you set `mod='prefix'`, you should also set param `absolut_ptn_len`"

    if flags is not None:
        rgx = re.compile(pattern, flags)
    else:
        rgx = re.compile(pattern)
    split_points = [0, ]
    n = len(text)
    while True:
        start = split_points[-1]
        if mod == "prefix" and len(split_points) > 1:
            start += absolut_ptn_len
        match = rgx.search(text[start:])

        if match:
            if mod == "prefix":
                split_points.append(match.span()[0] + start)
            elif mod == "suffix":
                split_points.append(match.span()[1] + start)
            elif mod == "full":
                split_points.append(match.span()[0] + start)
                split_points.append(match.span()[1] + start)
        else:
            break

    if len(split_points) > 1 and split_points[1] == 0:
        split_points = split_points[1:]
    if split_points[-1] != n:
        split_points.append(n)
    return [text[split_points[i]:split_points[i + 1]] for i in range(len(split_points) - 1)]


def get_credits(text):
    credits_ = []
    persons = []
    credit_last_match_pt = 0
    for credit in patterns_credit:
        flag = True
        for ptn in patterns_credit[credit]:
            if flag:
                rgx = re.compile(ptn, re.I)
                match = rgx.search(text)
                if match is not None:
                    credits_.append(credit)
                    flag = False
                    match_pt = match.span()[1]
                    credit_last_match_pt = max(credit_last_match_pt, match_pt)

    if credits_:
        # if there is semicolon, it's easy
        if re.search(pattern_semicolon, text) is not None:
            persons_maybe_text = re.split(pattern_semicolon, text, maxsplit=1)[1]
        # no semicolon, use `credit_last_match_pt`
        else:
            persons_maybe_text = text[credit_last_match_pt:]
    else:
        persons_maybe_text = text

    persons_maybe = re.split(r",|and|\\|\|", persons_maybe_text)
    for p in persons_maybe:
        person = rm_presuffix(r'\W|_', p)
        if person:
            persons.append(person)
    return [credits_, persons]


def get_disc(text):
    if re.search(pattern_disc, text) is not None:
        return int(re.search(r"\d+", text).group())
    return None


def get_tracks(text):
    tracks = []
    match = re.search(pattern_track, text)
    if match is not None and match.group() == text:
        text = re.sub(r"(?<=\d)\s*to\s*(?=\d)", '~', text)
        text = re.sub(r"(?<=\d)\s*-\s*(?=\d)", '~', text)
        text = re.sub(r"(?<=\d)\s*~\s*(?=\d)", '~', text)
        for num_str in re.split(r"[^0-9~]+", text):
            if num_str:
                if '~' in num_str:
                    num_str_split = num_str.split('~')
                    if len(num_str_split) == 2:
                        [start, end] = num_str_split
                        [start, end] = [int(start), int(end)]
                        tracks += list(range(start, end + 1))
                else:
                    tracks.append(int(num_str))
    return tracks


if __name__ == '__main__':
    print(split_by_re(pattern_track, "M. 01 3st"))



import requests
from pyvgmdb.utils import SafeEtree
from lxml import etree
import warnings


class Artist:
    def __init__(self, artist_id, proxies=None, headers=None):
        """
        :param artist_id: e.g. 9621, then we will parse: https://vgmdb.net/artist/9621
        :param proxies: if you need proxies to visit https://vgmdb.net, please set this param
        :param headers: not necessary
        """
        self.artist_id = str(artist_id)
        self.homepage_url = "https://vgmdb.net/artist/" + self.artist_id
        self.proxies = {'http': f'http://127.0.0.1:{proxies}', 'https': f'http://127.0.0.1:{proxies}'}
        self.headers = headers
        self.tree = self.getPageEtree()
        self.albums = self.getDiscography() if self.tree is not None else None

    def getPageEtree(self):
        args = {'url': self.homepage_url}
        if self.proxies:
            args['proxies'] = self.proxies
        if self.headers:
            args['headers'] = self.headers
        response = requests.get(**args)
        page_text = response.text
        if response.status_code == 404 or "Artist not found" in page_text:
            warnings.warn(f"can not find artist with id {self.artist_id}")
            return
        if "No albums found" in "".join(etree.HTML(page_text).xpath('//*[@id="albumlist"]/div[1]//text()')):
            warnings.warn(f"artist with id {self.artist_id} maybe have no album, please check again")
            return

        tree = SafeEtree(etree.HTML(page_text))
        return tree

    def getDiscography(self):
        discography_list = self.tree.xpath('discography', final=False)
        albums_all = []
        for discography in discography_list:
            albums_in_one_year = []
            year = discography.xpath('year')[0]
            tr_list = discography.xpath('tr', final=False)
            for i, tr in enumerate(tr_list):
                # first tr element is the year
                if i == 0:
                    continue
                date = tr.xpath('date')[0]
                album_title = tr.xpath('album_title')[0]
                album_url = tr.xpath('album_url')[0]
                # TODO: 'serve_as' list pretreatment
                serve_as = tr.xpath('serve_as')
                albums_in_one_year.append({
                    'date': date,
                    'album_title': album_title,
                    'album_url': album_url,
                    'serve_as': serve_as,
                })
            albums_all.append({'year': year, 'albums': albums_in_one_year, })
        return albums_all

    def get_albums_url_list(self):
        url_list = []
        for albums_in_one_year in self.albums:
            for album in albums_in_one_year['albums']:
                url_list.append(album['album_url'])
        return url_list

    def downloadPortrait(self, fp):
        """TODO"""

    def getBasicInfo(self):
        """TODO"""


if __name__ == '__main__':
    import numpy as np
    import time
    test_id = list(range(1, 10000))
    np.random.shuffle(test_id)
    bad_example = []
    for k in test_id[:50]:
        if k in bad_example:
            continue
        print("current artist id：", k)
        try:
            print(Artist(k, proxies=7890).albums)
        except Exception:
            bad_example.append(k)
            print("获取失败，请手动检查，id=", k)
        time.sleep(12)
    print(bad_example)



from threading import Thread
from typing import *


# 定义一个MyThread线程类， 可以获取线程的return返回值
class MyThread(Thread):
    def __init__(self, func, args=()):
        super(MyThread, self).__init__()
        self.func = func
        self.args = args

    def run(self):
        self.result = self.func(*self.args)

    def get_result(self):
        Thread.join(self)  # 等待线程执行完毕
        try:
            return self.result
        except Exception:
            print("线程返回结果失败")
            return None


# 并行任务模板，可以用于需要汇总最终结果的情况
# 如果函数返回的结果是列表，需要把内层列表解包时，则可设置：unpacked=True
def res_pool_parallel(func, args_list: List[Tuple], unpacked=False):
    res_pool = []
    thread_pool = []
    for args in args_list:
        my_thread = MyThread(
            func,
            args,
        )
        my_thread.start()
        thread_pool.append(my_thread)
    for thread in thread_pool:
        res = thread.get_result()
        if unpacked:
            res_pool += res
        else:
            res_pool.append(res)
    return res_pool



import warnings
from pyvgmdb.re_tools import *
import re


class Notes:
    def __init__(self, album_info):
        self.album_id = album_info['album_id']
        self.note_text = album_info['album_notes'].replace('&#160;', '').replace('&', ',')
        self.track_list = album_info['track_list_info'][0]['track_list']
        self.disc_num = len(self.track_list)
        self.album_note_parse = self.initAlbumNoteParser()
        self.parse()

    def initAlbumNoteParser(self):
        return [
            [
                {key: None for key in patterns_credit.keys()}
                for _ in range(len(self.track_list[disc]['disc_track_list']))
            ]
            for disc in range(self.disc_num)
        ]

    # write in album_note_parse
    def update(self, disc, credits_, tracks):
        for t in tracks:
            for c in credits_[0]:
                if disc <= self.disc_num and t <= len(self.album_note_parse[disc - 1]):
                    self.album_note_parse[disc - 1][t - 1][c] = credits_[1]
                else:
                    warnings.warn(f"disc num or track num index out of range in album "
                                  f"with id {self.album_id}")

    def parse(self):
        # not initialized
        cur_disc = 1

        # initialized every line
        cur_credits = [[], []]
        cur_tracks = []

        for item in self.note_text.split("<br /><br />"):
            line = []
            for text in re.split('<br />', item):
                text = text.replace('/', ',')
                text = re.sub(r"\d+(?=[a-zA-Z])", "", text)
                for _ in split_by_re(pattern_track, text, "full"):
                    line += split_by_re(pattern_disc, _, mod="full", flags=re.I)
            for text in line:
                tmp_disc = get_disc(text)
                if tmp_disc is None:
                    tmp_tracks = get_tracks(text)
                    if tmp_tracks:
                        tmp_credits = [[], []]
                    else:
                        tmp_credits = get_credits(text)
                else:
                    tmp_tracks = []
                    tmp_credits = [[], []]

                if tmp_disc is not None:
                    cur_disc = tmp_disc
                elif tmp_tracks:
                    cur_tracks = tmp_tracks
                    if cur_credits[0] and cur_credits[1]:
                        self.update(cur_disc, cur_credits, cur_tracks)
                        cur_credits[1] = []
                        cur_tracks = []
                    elif cur_credits[0] or cur_credits[1]:
                        warnings.warn(f"there is a new mod in album notes, "
                                      f"see details in album with id {self.album_id}")
                elif tmp_credits[0] and tmp_credits[1]:
                    cur_credits = tmp_credits
                    if cur_tracks:
                        self.update(cur_disc, cur_credits, cur_tracks)
                        cur_credits = [[], []]
                elif tmp_credits[0]:
                    cur_credits = tmp_credits
                elif tmp_credits[1]:
                    if cur_credits[0] and cur_tracks:
                        cur_credits[1] = tmp_credits[1]
                        self.update(cur_disc, cur_credits, cur_tracks)
                        cur_credits = [[], []]
                    elif cur_credits[0]:
                        cur_credits[1] = tmp_credits[1]
            cur_credits = [[], []]
            cur_tracks = []



main_xpaths_dict = {
    'xpath': '',

    # artist
    'artist_name_en': {'xpath': '//*[@id="innermain"]/span[2]/text()'},
    'artist_sex': {'xpath': '//*[@id="leftfloat"]/img/@alt'},
    'artist_name_jp': {'xpath': '//*[@id="leftfloat"]/span/text()'},
    'artist_img_url': {'xpath': '//*[@id="leftfloat"]/div[1]/a/@href'},
    'artist_left_bar': {'xpath': '//*[@id="leftfloat"]/div'},
    'artist_Notes': {'xpath': '//*[@id="rightfloat"]/div[2]/div/text()'},
    'artist_discography': {
        'xpath': '//*[@id="discotable"]/table/tbody',
        'year': {'xpath': './/h3[@class="time"]/text()'},
        # note: list index start from 1
        'tr': {
            'xpath': './tr',
            'date': {'xpath': './/td[@class="label"]/text()'},
            'album_title': {'xpath': './/a[contains(@class,"albumtitle")]/@title'},
            'album_url': {'xpath': './/a[contains(@class,"albumtitle")]/@href'},
            # 'serve_as' is a list, like: ['Composer', 'Arranger']
            'serve_as': {'xpath': './/span[contains(@class,"smallfont") and contains(@class,"label")]//text()'},
        }
    },

    # album
    'album_name_en': {'xpath': '//*[@id="innermain"]/h1/span[@lang="en"]/text()'},
    'album_name_ja': {'xpath': '//*[@id="innermain"]/h1/span[@lang="ja"]/text()'},
    'album_name_ja_latn': {'xpath': '//*[@id="innermain"]/h1/span[@lang="ja-Latn"]/text()'},
    'album_info_tr': {'xpath': '//div[@id="rightfloat"]//table[@id="album_infobit_large"]//tr'},
    'album_credits': {'xpath': '//div[@id="collapse_credits"]//table[@id="album_infobit_large"]//tr'},
    'track_list_span': {'xpath': '//*[@id="tracklist"]/span'},
}



from lxml import etree
from pyvgmdb.xpaths import main_xpaths_dict


class SafeEtree:
    def __init__(self, tree, dict_path=None):
        self.etree = tree
        self.dict_path = dict_path

    def xpath(self, key, final=True):
        xpath_dict = main_xpaths_dict
        if self.dict_path is not None:
            for p in self.dict_path:
                xpath_dict = xpath_dict[p]

        if key not in xpath_dict:
            raise Exception(f"key '{key}' is not found in dict path: {self.dict_path}"
                            f"please add xpath item in file: vgmdb.xpaths.py")
        xpath = xpath_dict[key]['xpath']
        res = self.etree.xpath(xpath)
        if not res:
            raise Exception(f"element named '{key}' is not found in current html page, "
                            f"maybe you can change its xpath in file: ./xpaths.py "
                            f"dict path is: {self.dict_path}")
        if final:
            return res
        else:
            new_dict_path = self.dict_path[:] if self.dict_path is not None else []
            new_dict_path.append(key)
            return self.__class__(res, new_dict_path)

    def __getitem__(self, item, final=False):
        return self.__class__(self.etree[item], self.dict_path)

    def __len__(self):
        return len(self.etree)

